/*
 *	File transfer protocol "TransIt"
 *		Version 2.00
 *	Copyright (C) 1987-1989 by ASCII Corporation.
 *
 *
 *	recv.c - receive files
 */
#include "transit.h"
#include <varargs.h>


static char
	recvbuff[MAXRECVWIN][LINESIZE];	/* buffer for data */
static int
	buffLength[MAXRECVWIN],		/* length of data in buffer */
	blockIndex = 0;			/* index */
static char
	buffActive[MAXRECVWIN];		/* flag:buffer active or not */
extern int
	blockAck;			/* block number to send acknowlige */

extern char *getenv();


/*
 *  send acknowlige
 */
sendBytes(length, type, va_alist)
int length;
char type;
va_dcl
{
	va_list ap;
	char *chp;
	char rawbuff[LINESIZE];
	char encbuff[LINESIZE];		/* buffer for encoded data */


	/* make block */
	rawbuff[DATALEN] = length;	/* number of data */
	if (type == DATABODY) {
		rawbuff[BLOCKNUM] = blockAck & 0xff;	/* block number */
	}
	else {
		rawbuff[BLOCKNUM] = 0;
	}
	rawbuff[BLOCKTYPE] = type;
	chp = &rawbuff[DATA];
	va_start(ap);
	while (length--) {
		*(chp++) = va_arg(ap, int);
	}
	va_end(ap);
	makeBlock(encbuff, rawbuff);
	writeBlock(encbuff);		/* send it */
}


/*
 *  receive block and send acknowlige
 */
static recvBlock(buffp)
char *buffp;
{
	register int i, length;
	int retry;


	if (buffActive[blockIndex] == TRUE) {
		/* block has been received */
		length = buffLength[blockIndex];
		bcopy(recvbuff[blockIndex], buffp, length + DATA);
		buffActive[blockIndex] = FALSE;
		if (++blockIndex >= MAXRECVWIN) {
			blockIndex = 0;
		}
		sendBytes(1, DATABODY, ACK);
		++blockAck;
		return (length);
	}

	for (retry = MAXRETRY; retry; --retry) {
		switch (length = readBlock(buffp)) {	/* receive one block */
		case TIMEOUT:	/* time out, retry */
			dispstat(TIMEOUT);
			break;
		case ERROR:	/* error */
			dispstat(ERROR);
			break;
		default:
			i = buffp[BLOCKNUM] & 0xff;
			i -= (blockAck & 0xff);
			if (i <= -0x80) {
				i += 0x100;
			}
			else if (i >= 0x80) {
				i -= 0x100;
			}
			if (i < 0) {
				--blockAck;
				sendBytes(1, DATABODY, ACK);
				++blockAck;
				continue;
			}
			if (i >= MAXRECVWIN) {
				break;	/* too large or already received */
			}

			if (i == 0) {	/* this block is it */
				if (++blockIndex >= MAXRECVWIN) {
					blockIndex = 0;
				}

				if (blockAck != 0) {	/* if not info block */
					sendBytes(1, DATABODY, ACK);
					++blockAck;
				}
				return (length);
			}
			if ((i += blockIndex) >= MAXRECVWIN) {
				i -= MAXRECVWIN;
			}
			bcopy(buffp, recvbuff[i], length + DATA);
			buffLength[i] = length;	/* save data length */
			buffActive[i] = TRUE;	/* set flag */
			++retry;		/* don't decrease counter */
			continue;
		}
		if (blockAck == 0) {	/* initiate block */
			sendBytes(4, CONTROL, WINDOW, MAXRECVWIN, GRPH, 1);
		}
		else {
			sendBytes(1, DATABODY, NAK);
		}
	}
	quit();			/* abort process */
}


/*
 *  receive initiate
 */
static recvInit()
{
	int i;


	blockAck = 0;
	for (i = MAXRECVWIN; i; --i) {
		buffActive[i] = 0;	/* clear flags */
	}
	sendBytes(4, CONTROL, WINDOW, MAXRECVWIN, GRPH, 1);
}


/*
 *  receive file
 */
recvFile(count, namepp)
int count;
char **namepp;
{
	register char *chp;
	FILE *fp;
	int length;
	long fsize;
	char rawbuff[LINESIZE], fname[LINESIZE];
	char buffer[LINESIZE];
	int textmode;


	if (count == 0) {
		count = -1;
	}
	for (; count; --count, ++namepp) {
		/* hand shake */
#ifdef DEBUG
		fputs("..Hand shaking", logptr);
		fflush(logptr);
#endif
		recvInit();
		if ((length = recvBlock(rawbuff)) == 0) {
			break;	/* null body */
		}

		chp = &rawbuff[DATA];
		if ((length = strlen(chp)) == 0) {
			break;	/* null name */
		}
		strcpy(fname, chp);	/* copy file name */

		chp += length + 1;
		sscanf(chp, "%ld", &fsize);	/* get file size */

		length = strlen(chp);
		chp += length + 1;

		textmode = FALSE;
		switch (*chp) {
		case TEXTMODE:
			textmode = TRUE;
		case BINARYMODE:
			chp++;
		}
		if (*chp == USEG1SET)
			gmode = 1;

		sendBytes(1, DATABODY, ACK);	/* send ACK to info block */
		++blockAck;

		if (count > 0) {/* no name specified */
			strcpy(fname, *namepp);	/* name specified */
		}
		else {
			for (chp = fname; *chp; ++chp) {
				if (!isalnum(*chp) && strchr(VALIDCHAR, *chp)
				    == NULL) {
					*chp = '\0';
					break;
				}
			}
		}
#ifdef DEBUG
		fprintf(logptr,"\r\"%s\" ", fname);
		fprintf(logptr,"(%s:%ldbytes) ", textmode ? "text" : "binary",
				fsize);
		fprintf(logptr,"-- %dbit encode\033[K\n", gmode == 1 ? 7 : 6);
#endif

		if (chp = getenv(DOWNDIR)) {
			strcpy(buffer, chp);
			length = strlen(buffer);
			if (strchr(PATHSPRTRS, buffer[length - 1]) == NULL) {
				strcat(buffer, "\\");
				length++;
			}
			strcat(buffer, fname);
			strcpy(fname, buffer);
		}

#if	0
		if (!oopt && access(fname, 0) != EOF) {
#ifdef DEBUG
			fprintf(logptr,
				"File already exists!\007 Over-write ?(y*/n) ");
#endif
			gets(buffer);
			if (*buffer == 'n' || *buffer == 'N') {
				do {
					printf("Enter new name ? ");
					gets(&fname[length]);
				} while (fname[length] == '\0');
			}
		}
#endif

		if (textmode)
			fp = fopen(fname, FOPENWRT);
		else
			fp = fopen(fname, FOPENWRB);
		if (fp == NULL) {
#ifdef DEBUG
			fprintf(logptr,"..Cannot open %s\n..Abort transfer\n",
				 fname);
#endif
			quit();
		}

		dispinit(fsize);

		while (TRUE) {
			if ((length = recvBlock(rawbuff)) == 0) {
				break;			/* null body */
			}
			if (rawbuff[BLOCKTYPE] == CONTROL) {
				if (rawbuff[DATA] == TERMINATE) {
					abort();	/* terminated */
				}
				continue;
			}
			if (rawbuff[BLOCKTYPE] != DATABODY) {
				continue;		/* ignore */
			}
			dispstat(SUCCESS, length);
			fwrite(rawbuff + DATA, 1, length, fp);
		}
		fclose(fp);
		dispdone();
	}
	if (count == 0) {	/* file names were specified */
		quit();		/* no more transfer */
	}
	sendBytes(1, DATABODY, ACK);		/* send ACK */
#ifdef DEBUG
	fputs("\r..End of transfer\033[K\007\007\n", logptr);
#endif
}
