/*
 *	File transfer protocol "TransIt"
 *		Version 2.00
 *	Copyright (C) 1987-1989 by ASCII Corporation.
 *
 *
 *	send.c - send files
 */
#include "transit.h"

static char
	encbuff[MAXSENDWIN][LINESIZE];	/* buffer for encoded data */
static int
	buffIndex = 0,			/* buffer index */
	blockNext = 0,			/* block number to send */
	windowWidth = MAXSENDWIN,	/* width of window */
	windowUsed = 0;			/* number of slot in use */
static int
	g1ok = 0;
extern int
	blockAck;			/* block number to wait ack */

extern char *getenv();


/*
 *  wait acknowlige
 */
static recvAck()
{
	int retry;
	char rawbuff[LINESIZE];
	register int i;


	for (retry = MAXRETRY; retry; --retry) {
		/* read acknowlige */
		i = readBlock(rawbuff);
		if (i != ERROR && i != TIMEOUT) {
			if (rawbuff[BLOCKTYPE] == CONTROL) {
				if (rawbuff[DATA] == TERMINATE) {
					abort();	/* terminate transfer */
				}
				continue;	/* ignore other control */
			}
			if (rawbuff[BLOCKTYPE] != DATABODY) {
				continue;	/* accepte data only */
			}
			i = rawbuff[BLOCKNUM] & 0xff;
			i -= (blockAck & 0xff);
			if (i <= -0x80) {
				i += 0x100;
			}
			else if (i >= 0x80) {
				i -= 0x100;
			}

			switch (rawbuff[DATA]) {
			case NAK:
				if (i == 0) {
					break;	/* no ACKed block */
				}
				--i;	/* decrease one */
				/* FALL down to ACK */
			case ACK:
				if (i < 0) {	/* already ACKed */
					continue;
				}
				if (i >= windowUsed) {
					break;	/* too large */
				}
				blockAck += (i + 1);
				windowUsed = blockNext - blockAck;
						/* more window */
				if (rawbuff[DATA] == ACK) {
					return;	/* ACK received */
				}
				break;
			}
			dispstat(NAK);
		}
		else if (i == TIMEOUT) {
			dispstat(TIMEOUT);
		}
		else {
			dispstat(ERROR);
		}
#ifdef DEBUG
		fflush(logptr);
#endif
		/* error or time out or NAK */
		if ((i = buffIndex - windowUsed) < 0) {
			i += MAXSENDWIN;	/* resend not ACKed block */
		}
		writeBlock(encbuff[i]);
	}
	quit();			/* too many retry */
}


/*
 *  send one block
 */
static sendBlock(buffp, count, type)
char *buffp, type;
int count;
{
	char rawbuff[LINESIZE];	/* buffer for raw data */


	/* make block */
	rawbuff[DATALEN] = count;	/* number of data */
	rawbuff[BLOCKNUM] = blockNext & 0xff;	/* block number */
	rawbuff[BLOCKTYPE] = type;	/* data type */
	bcopy(buffp, rawbuff + DATA, count);	/* copy data */
	makeBlock(encbuff[buffIndex], rawbuff);
	writeBlock(encbuff[buffIndex]);	/* send it */
	++blockNext;
	++windowUsed;
	if (++buffIndex >= MAXSENDWIN) {
		buffIndex = 0;
	}
}


/*
*	Get acknowledge block
*/
static WaitAck()
{
	int count;


	count = rsdRetSize();
	while (windowUsed == windowWidth || count > DATA + 1) {
		recvAck();
		count = rsdRetSize();
	}
}


/*
 *  send initiate
 */
static sendInit()
{
	int retry, i;
	char buffer[LINESIZE];
	register char *chp;


	blockAck = 0;
	blockNext = 0;
	for (retry = MAXRETRY; retry; --retry) {
		if ((i = readBlock(buffer)) == ERROR) {
			continue;	/* error, retry */
		}
		if (i == TIMEOUT) {
			continue;	/* error, retry */
		}
		if (buffer[BLOCKTYPE] != CONTROL) {
			continue;	/* accepte control only */
		}
		for (chp = buffer + DATA; i; ++chp, --i) {
			switch (*chp) {
			case TERMINATE:
				abort();
				break;
			case WINDOW:
				windowWidth = *(++chp);	/* get window width */
				if (windowWidth > MAXSENDWIN) {
					windowWidth = MAXSENDWIN;
				}
				break;
			case GRPH:
				g1ok = *(++chp);
				break;
			}
		}
		return;
	}
	quit();
}


/*
 *  send files
 */
sendFile(count, fnamepp)
int count;
char **fnamepp;
{
	FILE *fp;
	int i, length;
	char buffer[LINESIZE], *chp;
	char fname[LINESIZE];
	struct stat fbuff;
	int datasize = DATASIZE0;


	for (; count; --count, ++fnamepp) {
		if (access(*fnamepp, 0) == EOF) {
			if ((chp = getenv(UPDIR)) == NULL || chp[0] == '\0')
				continue;	/* ignore this file */
			strcpy(fname, chp);
			length = strlen(fname);
			if (strchr(PATHSPRTRS, fname[length - 1]) == NULL) {
				strcat(fname, "\\");
			}
			strcat(fname, *fnamepp);
		}
		else {
			strcpy(fname, *fnamepp);
		}

		/* open file */
		if (topt)
			fp = fopen(fname, FOPENRDT);
		else
			fp = fopen(fname, FOPENRDB);
		if (fp == NULL) {
			continue;	/* ignore this file */
		}

		/* get file info. */
		fstat(fileno(fp), &fbuff);

		/* remove path name prefix */
		while (chp = (char *) strpbrk(*fnamepp, PATHSPRTRS)) {
			*fnamepp = chp + 1;
		}

		/* wait request and receive window width */
#ifdef DEBUG
		fputs("..Hand shaking", logptr);
#endif
		sendInit();

#ifdef DEBUG
		fprintf(logptr,"\r\"%s\" ", *fnamepp);
		fprintf(logptr,"(%s:", topt ? "text" : "binary");
		fprintf(logptr,"%ldbytes) ", fbuff.st_size);
		fprintf(logptr,"-- %dbit encode\n", g1ok == 1 ? 7 : 6);
#endif

		dispinit(fbuff.st_size);

		/* send file info. */
		strcpy(buffer, *fnamepp);	/* copy file name */
		length = strlen(buffer) + 1;
		sprintf(&buffer[length], "%ld", fbuff.st_size);
						/* set file size */
		length += strlen(&buffer[length]) + 1;
		buffer[length++] = topt ? TEXTMODE : BINARYMODE;
		if (g1ok == 1)
			buffer[length++] = USEG1SET;

		sendBlock(buffer, length, DATABODY);	/* send file info */
		if (g1ok == 1) {
			gmode = 1;
			datasize = DATASIZE1;
		}
		WaitAck();				/* wait ACK block */

		/* send file body */
		do {
			length = fread(buffer, 1, datasize, fp);
			sendBlock(buffer, length, DATABODY);
			WaitAck();			/* wait ACK block */
			dispstat(SUCCESS, length);
		} while (length);
		while (windowUsed > 0) {
			recvAck();
		}
		fclose(fp);
		dispdone();
	}

#ifdef DEBUG
	fputs("..Hand shaking", logptr);
#endif
	sendInit();			/* initiate */
	sendBlock("", 0, DATABODY);	/* send null block */
	while (windowUsed > 0) {
		recvAck();
	}
#ifdef DEBUG
	fputs("\r..End of Transfer\007\007\033[K\n", logptr);
#endif
}
